==> jackrabbit-oak-1.90.0-src.zip: jackrabbit-oak-1.90.0/oak-segment-azure/src/main/java/org/apache/jackrabbit/oak/segment/azure/AzureSegmentArchiveWriter.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.jackrabbit.oak.segment.azure;

import com.azure.core.util.BinaryData;
import com.azure.core.util.Context;
import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.models.BlobStorageException;
import com.azure.storage.blob.options.BlockBlobSimpleUploadOptions;
import com.azure.storage.blob.specialized.BlockBlobClient;
import org.apache.jackrabbit.oak.commons.Buffer;
import org.apache.jackrabbit.oak.commons.time.Stopwatch;
import org.apache.jackrabbit.oak.segment.azure.util.Retrier;
import org.apache.jackrabbit.oak.segment.remote.AbstractRemoteSegmentArchiveWriter;
import org.apache.jackrabbit.oak.segment.remote.RemoteSegmentArchiveEntry;
import org.apache.jackrabbit.oak.segment.remote.WriteAccessController;
import org.apache.jackrabbit.oak.segment.spi.monitor.FileStoreMonitor;
import org.apache.jackrabbit.oak.segment.spi.monitor.IOMonitor;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.util.concurrent.TimeUnit;

import static org.apache.jackrabbit.oak.segment.azure.AzureUtilities.readBufferFully;
import static org.apache.jackrabbit.oak.segment.remote.RemoteUtilities.OFF_HEAP;
import static org.apache.jackrabbit.oak.segment.remote.RemoteUtilities.getSegmentFileName;

public class AzureSegmentArchiveWriter extends AbstractRemoteSegmentArchiveWriter {

    private final BlobContainerClient blobContainerClient;

    private final String archiveName;

    private final String archivePathPrefix;

    private final Retrier retrier = Retrier.withParams(
            Integer.getInteger("azure.segment.archive.writer.retries.max", 16),
            Integer.getInteger("azure.segment.archive.writer.retries.intervalMs", 5000)
    );

    public AzureSegmentArchiveWriter(BlobContainerClient blobContainerClient, String rootPrefix, String archiveName, IOMonitor ioMonitor, FileStoreMonitor monitor, WriteAccessController writeAccessController) {
        super(ioMonitor, monitor);
        this.blobContainerClient = blobContainerClient;
        this.archiveName = AzureUtilities.ensureNoTrailingSlash(archiveName);
        this.archivePathPrefix = AzureUtilities.asAzurePrefix(rootPrefix, archiveName);
        this.writeAccessController = writeAccessController;
        this.created = AzureUtilities.archiveExists(blobContainerClient, archivePathPrefix);
    }

    @Override
    public String getName() {
        return archiveName;
    }

    @Override
    protected void doWriteArchiveEntry(RemoteSegmentArchiveEntry indexEntry, byte[] data, int offset, int size) throws IOException {

        writeAccessController.checkWritingAllowed();

        long msb = indexEntry.getMsb();
        long lsb = indexEntry.getLsb();
        String segmentName = getSegmentFileName(indexEntry);
        BlockBlobClient blob = getBlockBlobClient(segmentName);
        ioMonitor.beforeSegmentWrite(new File(blob.getBlobName()), msb, lsb, size);
        Stopwatch stopwatch = Stopwatch.createStarted();
        try {
            // Upload the binary data and set its metadata using a single HTTP call,
            // overwriting an existing blob if necessary. Wrapping the byte array in a
            // ByteArrayInputStream avoids creating a copy of the data range.
            BinaryData binaryData = BinaryData.fromStream(new ByteArrayInputStream(data, offset, size), (long) size);
            BlockBlobSimpleUploadOptions options = new BlockBlobSimpleUploadOptions(binaryData)
                    .setMetadata(AzureBlobMetadata.toSegmentMetadata(indexEntry));
            blob.uploadWithResponse(options, null, Context.NONE);
        } catch (BlobStorageException e) {
            throw new IOException(e);
        }
        ioMonitor.afterSegmentWrite(new File(blob.getBlobName()), msb, lsb, size, stopwatch.elapsed(TimeUnit.NANOSECONDS));
    }

    @Override
    protected Buffer doReadArchiveEntry(RemoteSegmentArchiveEntry indexEntry)  throws IOException {
        Buffer buffer;
        if (OFF_HEAP) {
            buffer = Buffer.allocateDirect(indexEntry.getLength());
        } else {
            buffer = Buffer.allocate(indexEntry.getLength());
        }
        readBufferFully(getBlockBlobClient(getSegmentFileName(indexEntry)), buffer);
        return buffer;
    }

    @Override
    protected void doWriteDataFile(byte[] data, String extension) throws IOException {
        retrier.execute(() -> {
            try {
                writeAccessController.checkWritingAllowed();

                getBlockBlobClient(getName() + extension).upload(BinaryData.fromBytes(data), true);
            } catch (BlobStorageException e) {
                throw new IOException(e);
            }
        });
    }

    @Override
    protected void afterQueueClosed() throws IOException {
        retrier.execute(() -> {
            try {
                writeAccessController.checkWritingAllowed();

                getBlockBlobClient("closed").upload(BinaryData.fromBytes(new byte[0]), true);
            } catch (BlobStorageException e) {
                throw new IOException(e);
            }
        });
    }

    @Override
    protected void afterQueueFlushed() {
        // do nothing
    }

    private BlockBlobClient getBlockBlobClient(String name) throws IOException {
        String blobFullName = archivePathPrefix + name;
        try {
            return blobContainerClient.getBlobClient(blobFullName).getBlockBlobClient();
        } catch (BlobStorageException e) {
            throw new IOException(e);
        }
    }
}

==> jackrabbit-oak-1.92.0-src.zip: jackrabbit-oak-1.92.0/oak-segment-azure/src/main/java/org/apache/jackrabbit/oak/segment/azure/AzureSegmentArchiveWriter.java <==
/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.jackrabbit.oak.segment.azure;

import com.azure.core.util.BinaryData;
import com.azure.core.util.Context;
import com.azure.storage.blob.BlobContainerClient;
import com.azure.storage.blob.models.BlobStorageException;
import com.azure.storage.blob.options.BlockBlobSimpleUploadOptions;
import com.azure.storage.blob.specialized.BlockBlobClient;
import org.apache.jackrabbit.oak.commons.Buffer;
import org.apache.jackrabbit.oak.commons.time.Stopwatch;
import org.apache.jackrabbit.oak.segment.azure.util.Retrier;
import org.apache.jackrabbit.oak.segment.remote.AbstractRemoteSegmentArchiveWriter;
import org.apache.jackrabbit.oak.segment.remote.RemoteSegmentArchiveEntry;
import org.apache.jackrabbit.oak.segment.remote.WriteAccessController;
import org.apache.jackrabbit.oak.segment.spi.monitor.FileStoreMonitor;
import org.apache.jackrabbit.oak.segment.spi.monitor.IOMonitor;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.TimeUnit;

import static org.apache.jackrabbit.oak.segment.azure.AzureUtilities.readBufferFully;
import static org.apache.jackrabbit.oak.segment.remote.RemoteUtilities.OFF_HEAP;
import static org.apache.jackrabbit.oak.segment.remote.RemoteUtilities.getSegmentFileName;

public class AzureSegmentArchiveWriter extends AbstractRemoteSegmentArchiveWriter {

    private final BlobContainerClient blobContainerClient;

    private final String archiveName;

    private final String archivePathPrefix;

    private final Retrier retrier = Retrier.withParams(
            Integer.getInteger("azure.segment.archive.writer.retries.max", 16),
            Integer.getInteger("azure.segment.archive.writer.retries.intervalMs", 5000)
    );

    public AzureSegmentArchiveWriter(BlobContainerClient blobContainerClient, String rootPrefix, String archiveName, IOMonitor ioMonitor, FileStoreMonitor monitor, WriteAccessController writeAccessController) {
        super(ioMonitor, monitor);
        this.blobContainerClient = blobContainerClient;
        this.archiveName = AzureUtilities.ensureNoTrailingSlash(archiveName);
        this.archivePathPrefix = AzureUtilities.asAzurePrefix(rootPrefix, archiveName);
        this.writeAccessController = writeAccessController;
        this.created = AzureUtilities.archiveExists(blobContainerClient, archivePathPrefix);
    }

    @Override
    public String getName() {
        return archiveName;
    }

    @Override
    protected void doWriteArchiveEntry(RemoteSegmentArchiveEntry indexEntry, byte[] data, int offset, int size) throws IOException {

        writeAccessController.checkWritingAllowed();

        long msb = indexEntry.getMsb();
        long lsb = indexEntry.getLsb();
        String segmentName = getSegmentFileName(indexEntry);
        BlockBlobClient blob = getBlockBlobClient(segmentName);
        ioMonitor.beforeSegmentWrite(new File(blob.getBlobName()), msb, lsb, size);
        Stopwatch stopwatch = Stopwatch.createStarted();
        try {
            // Upload the binary data and set its metadata using a single HTTP call,
            // overwriting an existing blob if necessary. Wrapping the byte array in a
            // ByteArrayInputStream avoids creating a copy of the data range.
            // Note: OAK-12094 fixes an interesting regression in heap usage, which
            // should be read for context when making changes in the lines below.
            BlockBlobSimpleUploadOptions options = new BlockBlobSimpleUploadOptions(new ByteArrayInputStream(data, offset, size), size)
                    .setMetadata(AzureBlobMetadata.toSegmentMetadata(indexEntry));
            blob.uploadWithResponse(options, null, Context.NONE);
        } catch (BlobStorageException e) {
            throw new IOException(e);
        }
        ioMonitor.afterSegmentWrite(new File(blob.getBlobName()), msb, lsb, size, stopwatch.elapsed(TimeUnit.NANOSECONDS));
    }

    @Override
    protected Buffer doReadArchiveEntry(RemoteSegmentArchiveEntry indexEntry)  throws IOException {
        Buffer buffer;
        if (OFF_HEAP) {
            buffer = Buffer.allocateDirect(indexEntry.getLength());
        } else {
            buffer = Buffer.allocate(indexEntry.getLength());
        }
        readBufferFully(getBlockBlobClient(getSegmentFileName(indexEntry)), buffer);
        return buffer;
    }

    @Override
    protected void doWriteDataFile(byte[] data, String extension) throws IOException {
        retrier.execute(() -> {
            try {
                writeAccessController.checkWritingAllowed();

                getBlockBlobClient(getName() + extension).upload(BinaryData.fromBytes(data), true);
            } catch (BlobStorageException e) {
                throw new IOException(e);
            }
        });
    }

    @Override
    protected void afterQueueClosed() throws IOException {
        retrier.execute(() -> {
            try {
                writeAccessController.checkWritingAllowed();

                getBlockBlobClient("closed").upload(BinaryData.fromBytes(new byte[0]), true);
            } catch (BlobStorageException e) {
                throw new IOException(e);
            }
        });
    }

    @Override
    protected void afterQueueFlushed() {
        // do nothing
    }

    private BlockBlobClient getBlockBlobClient(String name) throws IOException {
        String blobFullName = archivePathPrefix + name;
        try {
            return blobContainerClient.getBlobClient(blobFullName).getBlockBlobClient();
        } catch (BlobStorageException e) {
            throw new IOException(e);
        }
    }
}
